/*
* gqviewarc.c -- archive handler for GQview
*
* Copyright (C) 2008  Kenny Wai-ip Lam
* waiiplam@yahoo.com.hk
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <time.h>
#include <sys/wait.h>

#define __USE_GNU
#include <string.h>
#include <dlfcn.h>

#define MAX_LEN 1000
#define ARC_ZIP 1
#define ARC_RAR 2

typedef struct Dir {
  struct Dir *next;
  DIR *dir;
  FILE *fp;
  pid_t pid;
  int arc;
  char path[MAX_LEN];
  off_t size;
  time_t t;
  int seq;
  int len;
  struct dirent ent;
  char ext[MAX_LEN - 256];
} Dir;

int (*_xstat)(int, const char *, struct stat *) = NULL;
int (*_access)(const char *, int) = NULL;
DIR *(*_opendir)(const char *) = NULL;
int (*_closedir)(DIR *) = NULL;
struct dirent *(*_readdir)(DIR *) = NULL;
int (*_open)(const char *, int, mode_t) = NULL;
int (*_close)(int) = NULL;

int debug = 0;
Dir *d_head = NULL;
int last_fd = -1;
pid_t last_pid;

void init() {
  unsetenv("LD_PRELOAD");

  _xstat = dlsym(RTLD_NEXT, "__xstat");
  _access = dlsym(RTLD_NEXT, "access");
  _opendir = dlsym(RTLD_NEXT, "opendir");
  _closedir = dlsym(RTLD_NEXT, "closedir");
  _readdir = dlsym(RTLD_NEXT, "readdir");
  _open = dlsym(RTLD_NEXT, "open");
  _close = dlsym(RTLD_NEXT, "close");
}

int stat_arc(int ver, char *s, struct stat *buf, char **p) {
  int a;

  if ((*p = strcasestr(s, ".zip/")) != NULL) a = ARC_ZIP;
  else if ((*p = strcasestr(s, ".rar/")) != NULL) a = ARC_RAR;
  else return 0;

  *p += 4;
  **p = '\0';

  if (_xstat == NULL) init();
  if (_xstat(ver, s, buf) == 0 && S_ISREG(buf->st_mode)) return a;
  else {
    **p = '/';
    return 0;
  }
}

int end_with(char *s1, int l1, char *s2, int l2) {
  //int l1 = strlen(s1), l2 = strlen(s2);
  char *p1, *p2;
  int i;

  if (l1 < l2) return 0;

  p1 = s1 + l1;
  p2 = s2 + l2;

  for (i = l2; i > 0; i--) {
    p1--;
    p2--;

    if (*p1 == *p2) continue;

    if (*p1 >= 'A' && *p1 <= 'Z') {
      if ((*p1 - 'A' + 'a') == *p2) continue;
    } /*else if (*p1 >= 'a' && *p1 <= 'z') {	// patten is always lower
      if ((*p1 - 'a' + 'A') == *p2) continue;
    }*/

    return 0;
  }

  return 1;
}

int arc_type(char *path) {
  int l = strlen(path);

  if (end_with(path, l, ".zip", 4)) return ARC_ZIP;
  if (end_with(path,  l,".rar", 4)) return ARC_RAR;
  return 0;
}

int open_proc(char *file, char *argv[], pid_t *pid) {
  int fd[2];

  if (pipe(fd) != 0) return -1;

  if (_close == NULL) init();

  while (waitpid(-1, NULL, WNOHANG) > 0);

  switch (*pid = vfork()) {
    case -1:

      _close(fd[0]);
      _close(fd[1]);
      return -1;

    case 0:

      _close(fd[0]);

      if (dup2(fd[1], 1) >= 0) {
	_close(fd[1]);
	execvp(file, argv);
      }

      exit(0);

    default:

      _close(fd[1]);
#ifdef DEBUG
      if (debug) fprintf(stderr, "%s: pid=%d\n", argv[0], *pid);
#endif
      return fd[0];
  }
}

void close1(FILE *fp, pid_t pid) {
  fclose(fp);
  waitpid(pid, NULL, 0);
}

int __xstat(int ver, const char *path, struct stat *buf) {
  char *p;
  Dir *dp;
  int r;

  if (stat_arc(ver, (char *)path, buf, &p) > 0) {
    for (dp = d_head; dp != NULL; dp = dp->next)
      if (strcmp(dp->path, path) == 0) {
	if (strcmp(dp->ent.d_name, p + 1) == 0) {
	  buf->st_size = dp->size;
	  buf->st_mtime = dp->t;
	}

	break;
      }

    *p = '/';
#ifdef DEBUG
    //if (debug) fprintf(stderr, "stat(\"%s\",)\n", path);
#endif
    return 0;
  }

  if (_xstat == NULL) init();
  r = _xstat(ver, (char *)path, buf);

  if (arc_type((char *)path) > 0 && r == 0 && S_ISREG(buf->st_mode)) {
#ifdef DEBUG
    //if (debug) fprintf(stderr, "stat(\"%s\",)\n", path);
#endif
    buf->st_mode ^= S_IFREG;
    buf->st_mode |= S_IFDIR;
  }

  return r;
}

int access(const char *pathname, int mode) {
  if (arc_type((char *)pathname) > 0) mode &= ~X_OK;

  if (_access == NULL) init();
  return _access(pathname, mode);
}

DIR *opendir(const char *name) {
  int a;
  struct stat buf;
  char s[MAX_LEN];
  char *argv[6];
  int fd;
  pid_t pid;
  FILE *fp;
  Dir *d;

  if (_opendir == NULL) init();

  if ((a = arc_type((char *)name)) > 0
	&& _xstat(_STAT_VER, (char *)name, &buf) == 0 && S_ISREG(buf.st_mode)) {
#ifdef DEBUG
    if (debug) fprintf(stderr, "opendir(\"%s\")\n", name);
#endif

    switch (a) {
      case ARC_ZIP:

	//snprintf(s, MAX_LEN, "unzip -lqq '%s'", name);
	argv[0] = "unzip";
	argv[1] = "-lqq";
	argv[2] = (char *)name;
	argv[3] = NULL;
	fd = open_proc("unzip", argv, &pid);

	break;
      default:	// rar

	//snprintf(s, MAX_LEN, "unrar v -c- -- '%s'", name);
	argv[0] = "unrar";
	argv[1] = "v";
	argv[2] = "-c-";
	argv[3] = "--";
	argv[4] = (char *)name;
	argv[5] = NULL;
	fd = open_proc("unrar", argv, &pid);
    }

    if (fd < 0) return NULL;

    fp = fdopen(fd, "r");

    if (fp == NULL) {
      _close(fd);
      waitpid(pid, NULL, 0);
      return NULL;
    }

    switch (a) {
      case ARC_RAR:

	for (;;) {
	  s[0] = '\0';
	  if (fscanf(fp, "%8[^\n\r]", s) == EOF) {
	    close1(fp, pid);
	    return NULL;
	  }

	  fscanf(fp, "%*[^\n\r]");
	  fscanf(fp, "%*[\n\r]");

	  if (strcmp(s, "--------") == 0) break;
	}
    }

    if ((d = (Dir *)malloc(sizeof(Dir))) == NULL) {
      close1(fp, pid);
      return NULL;
    }

    d->dir = _opendir("/tmp");
    d->fp = fp;
    d->pid = pid;
    d->arc = a;
    strncpy(d->path, name, MAX_LEN - 1);
    d->path[MAX_LEN - 1] = '\0';
    d->size = 0;
    d->t = 0;
    d->seq = 0;
    d->len = 0;
    d->ent.d_ino = 0;
    d->ent.d_off = 0;
    d->ent.d_reclen = sizeof(struct dirent) + MAX_LEN - 256;
    d->ent.d_type = DT_REG;
    d->ent.d_name[0] = '\0';

    d->next = d_head;
    d_head = d;

    return d->dir;
  }

  return _opendir(name);
}

int closedir(DIR *dir) {
  Dir *dp, *dp1 = NULL;

  for (dp = d_head; dp != NULL; dp = dp->next) {
    if (dp->dir == dir) {
#ifdef DEBUG
      if (debug) fprintf(stderr, "closedir(%p)\n", dir);
#endif
      if (dp->fp != NULL) close1(dp->fp, dp->pid);

      if (dp1 == NULL) d_head = dp->next;
      else dp1->next = dp->next;
      free(dp);

      break;
    }

    dp1 = dp;
  }

  if (_closedir == NULL) init();
  return _closedir(dir);
}

struct dirent *readdir(DIR *dir) {
  Dir *dp;
  unsigned int size;
  struct tm t;
  char name[MAX_LEN];
  unsigned int u;
  char c;
  char *p;
  int l;
  struct dirent *d;

  for (dp = d_head; dp != NULL; dp = dp->next)
    if (dp->dir == dir) {
      if (dp->fp == NULL) return NULL;

      for (;;) {
	switch (dp->arc) {

	  case ARC_ZIP:
	    if (fscanf(dp->fp, "%u", &size) != 1) {
	      close1(dp->fp, dp->pid);
	      dp->fp = NULL;
	      return NULL;
	    }

	    fscanf(dp->fp, "%u-", &u);
	    t.tm_mon = u - 1;
	    fscanf(dp->fp, "%u-", &u);
	    t.tm_mday = u;
	    fscanf(dp->fp, "%u", &u);
	    t.tm_year = u + 100;
	    fscanf(dp->fp, "%u:", &u);
	    t.tm_hour = u;
	    fscanf(dp->fp, "%u:%*u", &u);
	    t.tm_min = u;
	    t.tm_sec = 0;

	    fscanf(dp->fp, "%*3[ ]");

	    name[0] = '\0';
	    fscanf(dp->fp, "%999[^\n\r]", name);	// MAX_LEN - 1
	    fscanf(dp->fp, "%*[^\n\r]");

	    break;
	  default:	// rar

	    if (fscanf(dp->fp, "%c", &c) == EOF || c != ' ') {
	      close1(dp->fp, dp->pid);
	      dp->fp = NULL;
	      return NULL;
	    }

	    name[0] = '\0';
	    fscanf(dp->fp, "%999[^\n\r]", name);	// MAX_LEN - 1
	    fscanf(dp->fp, "%*[^\n\r]");

	    if (fscanf(dp->fp, "%u", &size) != 1) {
	      close1(dp->fp, dp->pid);
	      dp->fp = NULL;
	      return NULL;
	    }

	    fscanf(dp->fp, "%*s%*s");
	    fscanf(dp->fp, "%u-", &u);
	    t.tm_mday = u;
	    fscanf(dp->fp, "%u-", &u);
	    t.tm_mon = u - 1;
	    fscanf(dp->fp, "%u", &u);
	    t.tm_year = u + 100;
	    fscanf(dp->fp, "%u:", &u);
	    t.tm_hour = u;
	    fscanf(dp->fp, "%u:%*u", &u);
	    t.tm_min = u;
	    t.tm_sec = 0;

	    fscanf(dp->fp, "%*[^\n\r]");
	    fscanf(dp->fp, "%*[\n\r]");
	}

	if (size > 0 && name[0] != '\0') {
	  p = rindex(name, '/');
	  l = p == NULL ? 0 : p - name + 1;

	  if (dp->seq <= 0 || l != dp->len
		|| strncmp(name, dp->ent.d_name, l) != 0) {
	    dp->seq++;
	    dp->len = l;
	  }

	  dp->size = size;
	  dp->t = mktime(&t);
	  strncpy(dp->ent.d_name, name, l);
	  snprintf(dp->ent.d_name + l, MAX_LEN - l, "%d\t%s",
		dp->seq, name + l);
	  break;
	}
      }

      return &dp->ent;
    }

  if (_readdir == NULL) init();
  d = _readdir(dir);

  if (d != NULL) {
    if (arc_type(d->d_name) > 0 && d->d_type == DT_REG) d->d_type = DT_DIR;
  }

  return d;
}

int open(const char *pathname, int flags, mode_t mode) {
  int a;
  struct stat buf;
  char *p, *p1;
  int l;
  char name[MAX_LEN];
  char *argv[7];
  int fd;
  pid_t pid;

  if ((a = stat_arc(_STAT_VER, (char *)pathname, &buf, &p)) > 0) {
    p1 = rindex(p + 1, '/');

    if (p1 == NULL) {
      l = 0;
      p1 = p + 1;
    } else {
      l = p1 - p;	// p1 - (p + 1) + 1
      p1++;
    }

    p1 = index(p1, '\t');

    if (p1 == NULL) p1 = p + 1;
    else {
      if (l + strlen(p1 + 1) >= MAX_LEN) return -1;

      strncpy(name, p + 1, l);
      strcpy(name + l, p1 + 1);
      p1 = name;
    }

    switch (a) {
      case ARC_ZIP:

	argv[0] = "unzip";
	argv[1] = "-p";
	argv[2] = (char *)pathname;
	argv[3] = p1;
	argv[4] = NULL;
	fd = open_proc("unzip", argv, &pid);

	break;
      default:	// rar

	argv[0] = "unrar";
	argv[1] = "p";
	argv[2] = "-inul";
	argv[3] = "--";
	argv[4] = (char *)pathname;
	argv[5] = p1;
	argv[6] = NULL;
	fd = open_proc("unrar", argv, &pid);
    }

    *p = '/';

    if (fd >= 0) {
      last_fd = fd;
      last_pid = pid;
    }

    return fd;
  }

  if (_open == NULL) init();
  return _open(pathname, flags, mode);
}

int close(int fd) {
  int r;

  if (_close == NULL) init();
  r = _close(fd);

  if (fd == last_fd && fd >= 0) {
#ifdef DEBUG
    if (debug) fprintf(stderr, "close(%d)\n", fd);
#endif
    last_fd = -1;
    waitpid(last_pid, NULL, 0);
  }

  return r;
}
